/* LoongArch specific backend library.
 *
 * Copyright (C) 2022 OpenAnolis community LoongArch SIG.
 * 
 * Author: Liwei Ge <geliwei@openanolis.org>
 *
 */

/*
   This file is free software; you can redistribute it and/or modify
   it under the terms of either

     * the GNU Lesser General Public License as published by the Free
       Software Foundation; either version 3 of the License, or (at
       your option) any later version

   or

     * the GNU General Public License as published by the Free
       Software Foundation; either version 2 of the License, or (at
       your option) any later version

   or both in parallel, as here.

   elfutils is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received copies of the GNU General Public License and
   the GNU Lesser General Public License along with this program.  If
   not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <assert.h>
#include <dwarf.h>

#define BACKEND loongarch_
#include "libebl_CPU.h"

static int
dwarf_bytesize_aux (Dwarf_Die *die, Dwarf_Word *sizep)
{
  int bits;
  if (((bits = 8 * dwarf_bytesize (die)) < 0
       && (bits = dwarf_bitsize (die)) < 0)
      || bits % 8 != 0)
    return -1;

  *sizep = bits / 8;
  return 0;
}

/* r4 - r5, $a0 - $a1 Argument registers / return value registers */
static int
pass_by_ref (const Dwarf_Op **locp)
{
  static const Dwarf_Op loc[] = { { .atom = DW_OP_reg4 } };
  *locp = loc;

  return 1;
}

static int
pass_in_gpr (const Dwarf_Op **locp, Dwarf_Word size)
{
  static const Dwarf_Op loc[] =
  {
    { .atom = DW_OP_reg4 }, { .atom = DW_OP_piece, .number = 8 },
    { .atom = DW_OP_reg5 }, { .atom = DW_OP_piece, .number = 8 }
  };
  *locp = loc;

  return size <= 8 ? 1 : 4;
}

/* r32 - r33, $fa0 - $fa1 Argument registers / return value registers */
static int
pass_in_fpr (const Dwarf_Op **locp, Dwarf_Word size)
{
  static const Dwarf_Op loc[] =
  {
    { .atom = DW_OP_regx, .number = 32 },
    { .atom = DW_OP_piece, .number = 8 },
    { .atom = DW_OP_regx, .number = 33 },
    { .atom = DW_OP_piece, .number = 8 }
  };
  *locp = loc;

  return size <= 8 ? 1 : 4;
}

int
loongarch_return_value_location (Dwarf_Die *functypedie, const Dwarf_Op **locp)
{
  /* Start with the function's type, and get the DW_AT_type attribute,
     which is the type of the return value.  */
  Dwarf_Die typedie;
  int tag = dwarf_peeled_die_type (functypedie, &typedie);
  if (tag <= 0)
    return tag;

  Dwarf_Word size = (Dwarf_Word)-1;

  if (tag == DW_TAG_structure_type || tag == DW_TAG_union_type                        
      || tag == DW_TAG_class_type || tag == DW_TAG_array_type)
  { 
    if (dwarf_aggregate_size (&typedie, &size) < 0)                                 
      return -1;

    if (size > 16)                                                             
        return pass_by_ref(locp);
    else
        return pass_in_gpr(locp, size); 
  }

  if (tag == DW_TAG_base_type                                                         
      || tag == DW_TAG_pointer_type || tag == DW_TAG_ptr_to_member_type)
  {
    if (dwarf_bytesize_aux (&typedie, &size) < 0)
    {
      if (tag == DW_TAG_pointer_type || tag == DW_TAG_ptr_to_member_type)
        size = 8;
      else
        return -1;
    }

    Dwarf_Attribute attr_mem;
    if (tag == DW_TAG_base_type)
    {
      Dwarf_Word encoding;
      if (dwarf_formudata (dwarf_attr_integrate (&typedie, DW_AT_encoding, &attr_mem),
                          &encoding) != 0)
        return -1;

      switch (encoding)
      {
        case DW_ATE_boolean:
        case DW_ATE_signed:
        case DW_ATE_unsigned:
        case DW_ATE_unsigned_char:
        case DW_ATE_signed_char:
          return pass_in_gpr(locp, size);

        case DW_ATE_float:
          switch (size)
          {
            case 4: /* single */
            case 8: /* double */
              return pass_in_fpr (locp, size);
            case 16:
              return pass_by_ref (locp);
            default:
              return -2;
          }
        case DW_ATE_complex_float:
          switch (size)
          {
            case 8: /* single */
            case 16: /* double */
              return pass_in_fpr (locp, size);
            case 32:
              return pass_by_ref (locp);
            default:
              return -2;
          }
      }
    }
    else
      return pass_in_gpr(locp, size);
  }

  *locp = NULL;
  return 0;
}
